/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-string.h"
#include "my-time.h"

#include "config.h"
#include "db.h"
#include "eval_vm.h"
#include "execute.h"
#include "list.h"
#include "log.h"
#include "match.h"
#include "options.h"
#include "parse_cmd.h"
#include "random.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "tasks.h"
#include "utils.h"
#include "unparse.h"
#include "parser.h"

typedef struct forked_task {
    int		id;
    Program    *program;
    activation  a;
    Var        *rt_env;
    int		f_index;
    time_t	start_time;
} forked_task;

typedef struct suspended_task {
    vm 	   the_vm;
    time_t start_time; 
    int    id;
} suspended_task;

typedef struct {
    char       *string;
    int		length;
} input_task;

typedef struct task {
    struct task	       	       	       *next;
    task_kind			       	kind;
    union {
	input_task	input;
	forked_task	forked;
	suspended_task	suspended;
    }					t;
} task;

typedef struct tqueue {
    /*
     * A task queue can be in one of four possible states, depending upon the
     * values of its `player' and `connected' slots.
     *
     * When people first connect to the MOO, they are assigned negative player
     * object numbers.  Once they log in (either by creating a new player
     * object or by connecting to an existing one), the connection is changed
     * to refer to the appropriate, non-negative, player number.
     *
     * Input tasks for logged-in players (with non-negative `player' slots) are
     * interpreted as normal MOO commands.  Those for anonymous players (with
     * negative `player' slots) are treated specially: they are passed to a
     * particular verb on the system object.  If that verb returns a valid
     * player object number, then that is used as the new player number for the
     * connection.
     *
     * The `connected' field is true iff this queue has an associated open
     * network connection.  Queues without such connections may nonetheless
     * contain input tasks, presumably typed ahead before disconnection.  Input
     * tasks in unconnected queues are discarded.
     *
     * If an unconnected queue becomes empty, it is destroyed.
     */
    struct tqueue      *next, **prev;	/* prev only valid on idle_tqueues */
    Objid		player;
    int			connected;
    task	       *first_input, **last_input;
    int			total_input_length;
    int			last_input_task_id;
    int			input_suspended;
    task	       *first_bg, **last_bg;
    int			usage;		/* a kind of inverted priority */
    int			num_bg_tasks;   /* in either here or waiting_tasks */
    int			hold_input;	/* make input tasks wait for read() */
    char	       *output_prefix, *output_suffix;
    Stream	       *program_stream;
    Objid		program_object;
    int			program_verbnum;

    char 		reading; /* boolean */
    vm			reading_vm;
    int 		reading_task_id;
} tqueue;

#define INPUT_HIWAT	MAX_QUEUED_INPUT
#define INPUT_LOWAT	(INPUT_HIWAT / 2)

#define NO_USAGE	-1

int		current_task_id;
static tqueue  *idle_tqueues = 0, *active_tqueues = 0;
static task    *waiting_tasks = 0;  /* forked and suspended tasks */

#define GET_START_TIME(ttt) \
    (ttt->kind == TASK_FORKED \
     ? ttt->t.forked.start_time \
     : ttt->t.suspended.start_time)



static void
deactivate_tqueue(tqueue *tq)
{
    tq->usage = NO_USAGE;

    tq->next = idle_tqueues;
    tq->prev = &idle_tqueues;
    if (idle_tqueues)
	idle_tqueues->prev = &(tq->next);
    idle_tqueues = tq;
}

static void
activate_tqueue(tqueue *tq)
{
    tqueue    **qq = &active_tqueues;
    
    while (*qq  &&  (*qq)->usage <= tq->usage)
	qq = &((*qq)->next);
    
    tq->next = *qq;
    tq->prev = 0;
    *qq = tq;
}

static void
ensure_usage(tqueue *tq)
{
    if (tq->usage == NO_USAGE) {
	tq->usage = active_tqueues ? active_tqueues->usage : 0;

	/* Remove tq from idle_tqueues... */
	*(tq->prev) = tq->next;
	if (tq->next)
	    tq->next->prev = tq->prev;

	/* ...and put it on active_tqueues. */
	activate_tqueue(tq);
    }
}

static tqueue *
find_tqueue(Objid player, int create_if_not_found)
{
    tqueue     *tq;
    
    for (tq = active_tqueues; tq; tq = tq->next)
	if (tq->player == player)
	    return tq;

    for (tq = idle_tqueues; tq; tq = tq->next)
	if (tq->player == player)
	    return tq;

    if (!create_if_not_found)
	return 0;
    
    tq = (tqueue *) mymalloc(sizeof(tqueue), M_TASK);

    deactivate_tqueue(tq);

    tq->player = player;
    tq->connected = 0;

    tq->first_input = tq->first_bg = 0;
    tq->last_input = &(tq->first_input);
    tq->last_bg = &(tq->first_bg);
    tq->num_bg_tasks = 0;
    tq->total_input_length = tq->input_suspended = 0;
    tq->last_input_task_id = 0;

    tq->output_prefix = tq->output_suffix = 0;
    tq->program_stream = 0;

    tq->reading = 0;
    tq->hold_input = 0;

    return tq;
}

static void
free_tqueue(tqueue *tq)
{
    /* Precondition: tq is on idle_tqueues */

    if (tq->output_prefix)
	free_str(tq->output_prefix);
    if (tq->output_suffix)
	free_str(tq->output_suffix);
    if (tq->program_stream)
	free_stream(tq->program_stream);
    if (tq->reading)
	free_vm(tq->reading_vm, 1);

    *(tq->prev) = tq->next;
    if (tq->next)
	tq->next->prev = tq->prev;

    myfree(tq, M_TASK);
}

static void
enqueue_bg_task(tqueue *tq, task *t)
{
    *(tq->last_bg) = t;
    tq->last_bg = &(t->next);
    t->next = 0;
}

static task *
dequeue_bg_task(tqueue *tq)
{
    task       *t = tq->first_bg;

    if (t) {
	tq->first_bg = t->next;
	if (t->next == 0)
	    tq->last_bg = &(tq->first_bg);
	else
	    t->next = 0;
	tq->num_bg_tasks--;
#if 0
	logf("DEQUEUE_BG_TASK: decr #%d => %d\n",
	     tq->player, tq->num_bg_tasks);
#endif
    }

    return t;
}

static task *
dequeue_input_task(tqueue *tq)
{
    task       *t = tq->first_input;

    if (t) {
	tq->first_input = t->next;
	if (t->next == 0)
	    tq->last_input = &(tq->first_input);
	else
	    t->next = 0;
	tq->total_input_length -= t->t.input.length;
	if (tq->input_suspended
	    && tq->connected
	    && tq->total_input_length < INPUT_LOWAT) {
	    server_resume_input(tq->player);
	    tq->input_suspended = 0;
	}
    }

    return t;
}

static task *
dequeue_any_task(tqueue *tq)
{
    task       *t = dequeue_input_task(tq);

    if (t)
	return t;
    else
	return dequeue_bg_task(tq);
}

static void
free_task(task *t, int strong)
{	/* for FORKED tasks, strong == 1 means free the rt_env also.
	   for SUSPENDED tasks, strong == 1 means free the vm also. */
    switch (t->kind) {
      case TASK_INPUT:
	free_str(t->t.input.string);
	break;
      case TASK_FORKED:
	if (strong)
	    free_rt_env(t->t.forked.rt_env, 
			t->t.forked.program->num_var_names);
	free_program(t->t.forked.program);
	break;
      case TASK_SUSPENDED:
	if (strong)
	    free_vm(t->t.suspended.the_vm, 1);	    
	break;
    }
    myfree(t, M_TASK);
}

static int
new_task_id(void)
{
    int	i;

    do {
	i = RANDOM();
    } while (i == 0);

    return i;
}

static void
start_programming(tqueue *tq, const char *argstr)
{
    Objid	player = tq->player;
    verbref_info vr_i;

    vr_i = get_verbref_info(player, argstr);
    
    if (vr_i.oid < 0)
	object_match_failed(player, vr_i.oid, argstr);
    else if (vr_i.verbnum == NOTHING)
	notify(player, "You must specify a verb; use the format object:verb.");
    else if (vr_i.ptr == 0)
	notify(player, "That object does not have that verb definition.");
    else if (!check_verbref_perms(vr_i, player)
	     || (server_flag_option("protect_set_verb_code")
		 && !is_wizard(player)))
	notify(player, "Permission denied.");
    else {
	Stream *str = new_stream(100);
	
	stream_printf(str, "Now programming %s:%s(%d).  Use \".\" to end.",
		      obj_name(vr_i.oid), verbref_name(vr_i), vr_i.verbnum);
	notify(player, reset_stream(str));
	tq->program_stream = str;
	tq->program_object = vr_i.oid;
	tq->program_verbnum = vr_i.verbnum;
    }
}

struct state {
    Objid	player;
    int		nerrors;
    char       *input;
};

static void
my_error(void *data, const char *msg)
{
    struct state       *s = data;

    notify(s->player, msg);
    s->nerrors++;
}

static void
my_ungetc(void *data, char c)
{
    struct state       *s = data;

    s->input--;
}

static int
my_getc(void *data)
{
    struct state       *s = data;

    if (*(s->input) != '\0')
	return *(s->input++);
    else
	return EOF;
}

static Parser_Client	client = { my_error, my_ungetc, my_getc };

static void
end_programming(tqueue *tq)
{
    Objid	player = tq->player;

    if (!valid(tq->program_object))
	notify(player, "That object appears to have disappeared ...");
    else {
	verbdef_info    v_i;

	v_i = verbdef_info_from_num(tq->program_object, tq->program_verbnum);
	
	if (!v_i.ptr)
	    notify(player, "That verb appears to have disappeared ...");
	else {
	    struct state s;
	    Program    *program;
	    char	buf[30];

	    s.player = tq->player;
	    s.nerrors = 0;
	    s.input = stream_contents(tq->program_stream);
	    
	    program = parse_program(client, (void *) &s);
	    
	    sprintf(buf, "%d error(s).", s.nerrors);
	    notify(player, buf);
	    
	    if (program) {
		if (program_from_info(v_i))
		    free_program(program_from_info(v_i));
		set_program_in_info(v_i, program);
		notify(player, "Verb programmed.");
	    } else
		notify(player, "Verb not programmed.");
	}
    }

    free_stream(tq->program_stream);
    tq->program_stream = 0;
}

static void
set_delimiter(char **slot, const char *string)
{
    if (*slot)
	free_str(*slot);
    if (*string == '\0')
	*slot = 0;
    else
	*slot = str_dup(string);
}

static int
do_command_task(tqueue *tq, char *command)
{
    if (tq->program_stream) {		/* We're programming */
	if (strcmp(command, ".") == 0)  /* Done programming */
	    end_programming(tq);
	else
	    stream_printf(tq->program_stream, "%s\n", command);
    } else {
	Parse_Info	pi;
	Program	       *p;
	int		junk;
	Objid		location = obj_location(tq->player);

	init_pi(&pi);
	pi.next = 0;
	pi.player = pi.caller = tq->player;

	if (!parse_command(command, &pi))
	    return 0;

	if (is_programmer(pi.player)  &&  verbcasecmp(".pr*ogram", pi.verb)) {
	    if (pi.args.v.list[0].v.num != 1)
		notify(pi.player, "Usage:  .program object:verb");
	    else
		start_programming(tq, pi.args.v.list[1].v.str);
	} else if (strcmp(pi.verb, "PREFIX") == 0)
	    set_delimiter(&(tq->output_prefix), pi.argstr);
	else if (strcmp(pi.verb, "SUFFIX") == 0)
	    set_delimiter(&(tq->output_suffix), pi.argstr);
	else {
	    Var	result, args;

	    result.type = TYPE_NUM; /* for free_var() if task isn't DONE */
	    junk = ((p = find_verb_on(pi.player, pi.player, pi.verb, &pi, 1))
		    || (p = find_verb_on(location, location, pi.verb, &pi, 1))
		    || (p = find_verb_on(pi.dobj, pi.dobj, pi.verb, &pi, 1))
		    || (p = find_verb_on(pi.iobj, pi.iobj, pi.verb, &pi, 1))
		    || (p = find_verb_on(location, location, "huh", &pi, 0)));
	
	    if (tq->output_prefix)
		notify(pi.player, tq->output_prefix);
	    args = parse_into_wordlist(command);
	    if (run_server_task_setting_id(pi.player, SYSTEM_OBJECT,
					   "do_command", args, command,
					   &result, &(tq->last_input_task_id))
		!= OUTCOME_DONE
		|| is_true(result)) {
		/* Do nothing more; we assume :do_command handled it. */
	    } else if (p) {
		tq->last_input_task_id = current_task_id = new_task_id();
		do_input_task(p, &pi, 0);
	    } else
		notify(pi.player, "I couldn't understand that.");
	    if (tq->output_suffix)
		notify(pi.player, tq->output_suffix);
	    free_var(result);
	}

	free_pi(&pi, 0);
    }

    return 1;
}

static int
do_login_task(tqueue *tq, char *command)
{
    Var		result;
    Var		args;
    Objid	old_max_object = db_size() - 1;

    result.type = TYPE_NUM;	/* In case #0:do_login_command does not exist
				 * or does not immediately return.
				 */

    args = parse_into_wordlist(command);
    run_server_task_setting_id(tq->player, SYSTEM_OBJECT, "do_login_command",
			       args, command, &result,
			       &(tq->last_input_task_id));
    if (result.type == TYPE_OBJ  &&  is_player(result.v.obj)) {
	Objid	new_player = result.v.obj;
	Objid	old_player = tq->player;
	tqueue *dead_tq = find_tqueue(new_player, 0);
	task   *t;

	tq->player = new_player;
	if (tq->num_bg_tasks) {
	    /* Cute; this un-logged-in connection has some queued tasks!
	     * Must copy them over to their own tqueue for accounting...
	     */
	    tqueue     *old_tq = find_tqueue(old_player, 1);

#if 0
	    logf("DO_LOGIN_TASK: copy %d for #%d\n",
		 tq->num_bg_tasks, old_player);
#endif
	    old_tq->num_bg_tasks = tq->num_bg_tasks;
	    while ((t = dequeue_bg_task(tq)) != 0)
		enqueue_bg_task(old_tq, t);
	    tq->num_bg_tasks = 0;
	}

	if (dead_tq) {		/* Copy over tasks from old queue for player */
#if 0
	    logf("DO_LOGIN_TASK: copy %d for #%d\n",
		 dead_tq->num_bg_tasks, tq->player);
#endif
	    tq->num_bg_tasks = dead_tq->num_bg_tasks;
	    while ((t = dequeue_any_task(dead_tq)) != 0) {
		if (t->kind == TASK_INPUT)
		    free_task(t, 0);
		else /* FORKED or SUSPENDED */
		    enqueue_bg_task(tq, t);
	    }
	    dead_tq->player = NOTHING; /* it'll be freed by run_ready_tasks */
	    dead_tq->num_bg_tasks = 0;
	}

	player_connected(old_player, new_player, new_player > old_max_object);
    }

    free_var(result);
    return 1;
}

static void
do_out_of_band_command(tqueue *tq, char *command)
{
    run_server_task(tq->player, SYSTEM_OBJECT, "do_out_of_band_command",
		    parse_into_wordlist(command), command, 0);
}

static int
is_out_of_input(tqueue *tq)
{
    return !tq->connected  &&  !tq->first_input;
}

/*
 * Exported interface
 */

task_queue
new_task_queue(Objid player)
{
    task_queue	result;
    tqueue     *tq = result.ptr = find_tqueue(player, 1);

    tq->connected = 1;

    return result;
}

void
free_task_queue(task_queue q)
{
    tqueue     *tq = (tqueue *) q.ptr;

    tq->connected = 0;

    /* Activate this queue to ensure that the pending read()ing task will
     * eventually get resumed, even if there's no more input for it.
     */

    if (tq->reading)
	ensure_usage(tq);
}

void
set_input_task_holding(task_queue q, int hold)
{
    tqueue     *tq = (tqueue *) q.ptr;

    tq->hold_input = hold;
    if (!hold  &&  tq->first_input) /* Anything to be done now? */
	ensure_usage(tq);
}

void
enqueue_input_task(task_queue q, const char *input)
{
    tqueue     *tq = (tqueue *) q.ptr;
    task       *t = (task *) mymalloc(sizeof(task), M_TASK);

    t->kind = TASK_INPUT;
    t->t.input.string = str_dup(input);
    tq->total_input_length += (t->t.input.length = strlen(input));

    *(tq->last_input) = t;
    tq->last_input = &(t->next);
    t->next = 0;

    if (!tq->hold_input  ||  tq->reading) /* Anything to do with this line? */
	ensure_usage(tq);

    if (!tq->input_suspended  &&  tq->total_input_length > INPUT_HIWAT) {
	server_suspend_input(tq->player);
	tq->input_suspended = 1;
    }
}

static void
enqueue_waiting(task *t)
{ /* either FORKED or SUSPENDED */

    time_t 	start_time = GET_START_TIME(t);
    Objid 	progr = (t->kind == TASK_FORKED
			 ? t->t.forked.a.progr
			 : progr_of_cur_verb(t->t.suspended.the_vm));
    tqueue     *tq = find_tqueue(progr, 1);

    tq->num_bg_tasks++;
#if 0
    logf("ENQUEUE_WAITING: incr #%d => %d\n", tq->player, tq->num_bg_tasks);
#endif
    if (!waiting_tasks  ||  start_time < GET_START_TIME(waiting_tasks)) {
	t->next = waiting_tasks;
	waiting_tasks = t;
    } else {
	task   *tt;
	
	for (tt = waiting_tasks; tt->next; tt = tt->next)
	    if (start_time < GET_START_TIME(tt->next))
		break;
	t->next = tt->next;
	tt->next = t;
    }	
}	

static void
enqueue_ft(Program *program, activation a, Var *rt_env,
	   unsigned f_index, time_t start_time, int id)
{
    task       *t = (task *) mymalloc(sizeof(task), M_TASK);

    t->kind = TASK_FORKED;
    t->t.forked.program = program;
    t->t.forked.a = a;
    t->t.forked.a.verb = str_ref(t->t.forked.a.verb);
    t->t.forked.a.verbname = str_ref(t->t.forked.a.verbname);
    t->t.forked.rt_env = rt_env;
    t->t.forked.f_index = f_index;
    t->t.forked.start_time = start_time;
    t->t.forked.id = id;
    
    enqueue_waiting(t);
}

static int
check_user_task_limit(Objid user)
{
    tqueue     *tq = find_tqueue(user, 0);
    int		limit = -1;
    Var		v;

    if (valid(user)
	&& get_prop_safe(user, "queued_task_limit", &v, 0) == E_NONE) {
	if (v.type == TYPE_NUM)
	    limit = v.v.num;
	free_var(v);
    }

    if (limit < 0)
	limit = server_int_option("queued_task_limit", -1);

    if (tq  &&  limit >= 0  &&  tq->num_bg_tasks >= limit)
	return 0;
    else
	return 1;
}

Var
enqueue_forked_task(Program *program, activation a, Var *rt_env, 
		    unsigned f_index, unsigned after_seconds)
{
    int		id = new_task_id();
    Var		v;

    if (check_user_task_limit(a.progr)) {
	enqueue_ft(program, a, rt_env, f_index, time(0) + after_seconds, id);
	v.type = TYPE_NUM;
	v.v.num = id;
    } else {
	v.type = TYPE_ERR;
	v.v.err = E_QUOTA;
    }

    return v;
}

enum error
enqueue_suspended_task(vm the_vm, void *data)
{
    int		after_seconds = *((int *) data);
    task       *t;

    if (check_user_task_limit(progr_of_cur_verb(the_vm))) {
	t = (task *) mymalloc(sizeof(task), M_TASK);
	t->kind = TASK_SUSPENDED;
	t->t.suspended.the_vm = the_vm;
	t->t.suspended.start_time = time(0) + after_seconds;
	t->t.suspended.id = current_task_id; /* remember old id */
    
	enqueue_waiting(t);
	return E_NONE;
    } else
	return E_QUOTA;
}

Var
read_input_now(Objid connection)
{
    tqueue     *tq = find_tqueue(connection, 0);
    task       *t;
    Var		r;

    if (!tq  ||  is_out_of_input(tq)) {
	r.type = TYPE_ERR;
	r.v.err = E_INVARG;
    } else if (!(t = dequeue_input_task(tq))) {
	r.type = TYPE_NUM;
	r.v.num = 0;
    } else {
	r.type = TYPE_STR;
	r.v.str = t->t.input.string;
	myfree(t, M_TASK);
    }

    return r;
}

enum error
make_reading_task(vm the_vm, void *data)
{
    Objid	player = *((Objid *) data);
    tqueue     *tq = find_tqueue(player, 0);
    
    if (!tq  ||  tq->reading  ||  is_out_of_input(tq))
	return E_INVARG;
    else {
	tq->reading = 1;
	tq->reading_vm = the_vm;
	tq->reading_task_id = current_task_id;
	if (tq->first_input)	/* Anything to read? */
	    ensure_usage(tq);
	return E_NONE;
    }
}

int
last_input_task_id(Objid player)
{
    tqueue     *tq = find_tqueue(player, 0);

    return tq ? tq->last_input_task_id : 0;
}

int
next_task_start(void)
{
    tqueue     *tq;

    for (tq = active_tqueues; tq; tq = tq->next)
	if (tq->first_input != 0  ||  tq->first_bg != 0)
	    return 0;

    if (waiting_tasks != 0) {
	int wait = (waiting_tasks->kind == TASK_FORKED
		    ? waiting_tasks->t.forked.start_time 
		    : waiting_tasks->t.suspended.start_time) - time(0);
	return (wait >= 0) ? wait : 0;
    } 

    else return -1;
}  

void
run_ready_tasks(void)
{
    task       *t, *next_t;
    time_t	now = time(0);
    tqueue     *tq, *next_tq;

    for (t = waiting_tasks; t && GET_START_TIME(t) <= now; t = next_t) {
	Objid progr = (t->kind == TASK_FORKED
		       ? t->t.forked.a.progr
		       : progr_of_cur_verb(t->t.suspended.the_vm)); 
	tqueue	*tq = find_tqueue(progr, 1);

	next_t = t->next;
	ensure_usage(tq);
	enqueue_bg_task(tq, t);
    }
    waiting_tasks = t;
    
    {
	int	did_one = 0;
	time_t	start = time(0);
	static char oob_prefix[] = OUT_OF_BAND_PREFIX;
	
	while (active_tqueues  &&  !did_one) {
	    /* Loop over tqueues, looking for a task */
	    tq = active_tqueues;

	    if (tq->reading && is_out_of_input(tq)) {
		tq->reading = 0;
		current_task_id = tq->reading_task_id;
		resume_reading_vm(tq->reading_vm, 0, 0);
		did_one = 1;
	    }
	    
	    while (!did_one) { /* Loop over tasks, looking for runnable one */
		if (tq->hold_input && !tq->reading)
		    t = dequeue_bg_task(tq);
		else
		    t = dequeue_any_task(tq);
		if (!t)
		    break;

		switch (t->kind) {
		  case TASK_INPUT:
		    if (sizeof(oob_prefix) > 1
			&&  !strncmp(oob_prefix, t->t.input.string,
				     sizeof(oob_prefix) - 1)) {
			do_out_of_band_command(tq, t->t.input.string);
			did_one = 1;
		    } else if (tq->reading) {
			tq->reading = 0; 
			current_task_id = tq->reading_task_id;
			resume_reading_vm(tq->reading_vm, t->t.input.string,
					  0);
			did_one = 1;
		    } else if (tq->connected) {
			add_command_to_history(tq->player, t->t.input.string);
			did_one = (tq->player >= 0
				   ? do_command_task
				   : do_login_task)(tq, t->t.input.string);
		    }
		    break;
		  case TASK_FORKED:
		    {
			forked_task 	ft;
			
			ft = t->t.forked;
			current_task_id = ft.id;
			do_forked_task(ft.program, ft.rt_env, ft.a,
				       ft.f_index, 0);
			did_one = 1;
		    }
		    break; 
		  case TASK_SUSPENDED:
		    current_task_id = t->t.suspended.id;
		    resume_suspended_task(t->t.suspended.the_vm, 0);
		    did_one = 1;
		    break;
		}
		free_task(t, 0);
	    }

	    active_tqueues = tq->next;

	    if (did_one) {
		/* Bump the usage level of this tqueue */
		time_t	end = time(0);
		
		tq->usage += end - start;
		activate_tqueue(tq);
	    } else {
		/* There was nothing to do on this tqueue, so deactivate it */
		deactivate_tqueue(tq);
	    }
	}
    }

    /* Free any unconnected and empty tqueues */
    for (tq = idle_tqueues; tq; tq = next_tq) {
	next_tq = tq->next;
	
	if (!tq->connected  &&  !tq->first_input  &&  tq->num_bg_tasks == 0)
	    free_tqueue(tq);
    }
}

enum outcome
run_server_task(Objid player, Objid what, const char *verb, Var args,
		const char *argstr, Var *result)
{
    return run_server_task_setting_id(player, what, verb, args, argstr, result,
				      0);
}

enum outcome
run_server_task_setting_id(Objid player, Objid what, const char *verb,
			   Var args, const char *argstr, Var *result,
			   int *task_id)
{
    verbdef_info v_i;
    
    v_i = get_verbdef_info(what, verb);
    if (verb_exists(v_i) && program_from_info(v_i)) {
	current_task_id = new_task_id();
	if (task_id)
	    *task_id = current_task_id;
	return do_server_task(what, verb, args, v_i.vloc,
			      name_from_info(v_i), program_from_info(v_i),
			      owner_from_info(v_i),
			      perms_from_info(v_i) & (1<<VP_DEBUG), player,
			      argstr, result);
    } else {
	/* simulate an empty verb */
	if (result) {
	    result->type = TYPE_NUM;
	    result->v.num = 0;
	}
	free_var(args);
	return OUTCOME_DONE;
    }
}

static Var
list_for_forked_task(forked_task ft) 
{
    Var		list;

    list = new_list(9);
    list.v.list[1].type = TYPE_NUM;
    list.v.list[1].v.num = ft.id;
    list.v.list[2].type = TYPE_NUM;
    list.v.list[2].v.num = ft.start_time;
    list.v.list[3].type = TYPE_NUM;
    list.v.list[3].v.num = 0;		/* OBSOLETE: was clock ID */
    list.v.list[4].type = TYPE_NUM;
    list.v.list[4].v.num = DEFAULT_BG_TICKS;	/* OBSOLETE: was clock ticks */
    list.v.list[5].type = TYPE_OBJ;
    list.v.list[5].v.obj = ft.a.progr;
    list.v.list[6].type = TYPE_OBJ;
    list.v.list[6].v.obj = ft.a.vloc;
    list.v.list[7].type = TYPE_STR;
    list.v.list[7].v.str = str_ref(ft.a.verbname);
    list.v.list[8].type = TYPE_NUM;
    list.v.list[8].v.num = get_lineno(ft.program, 0, ft.f_index);
    list.v.list[9].type = TYPE_OBJ;	
    list.v.list[9].v.obj = ft.a.this;

    return list;
}

static Var
list_for_vm(Var partial_list, vm the_vm)
{ /* code factoring: list_for_suspended_task and list_for_reading_task */
    partial_list.v.list[5].type = TYPE_OBJ;
    partial_list.v.list[5].v.obj = progr_of_cur_verb(the_vm);
    partial_list.v.list[6].type = TYPE_OBJ;
    partial_list.v.list[6].v.obj = top_activ(the_vm).vloc;
    partial_list.v.list[7].type = TYPE_STR;
    partial_list.v.list[7].v.str = str_ref(top_activ(the_vm).verbname);
    partial_list.v.list[8].type = TYPE_NUM;
    partial_list.v.list[8].v.num = suspended_lineno_of_vm(the_vm);
    partial_list.v.list[9].type = TYPE_OBJ;	
    partial_list.v.list[9].v.obj = top_activ(the_vm).this;

    return partial_list;
}

static Var
list_for_suspended_task(suspended_task st)
{
    Var		list;

    list = new_list(9);
    list.v.list[1].type = TYPE_NUM;
    list.v.list[1].v.num = st.id;
    list.v.list[2].type = TYPE_NUM;
    list.v.list[2].v.num = st.start_time;
    list.v.list[3].type = TYPE_NUM;
    list.v.list[3].v.num = 0;		/* OBSOLETE: was clock ID */
    list.v.list[4].type = TYPE_NUM;
    list.v.list[4].v.num = DEFAULT_BG_TICKS;	/* OBSOLETE: was clock ticks */

    return list_for_vm(list, st.the_vm);
}

static Var
list_for_reading_task(Objid player, int reading_task_id, vm the_vm)
{
    Var		list;

    list = new_list(9);
    list.v.list[1].type = TYPE_NUM;
    list.v.list[1].v.num = reading_task_id;
    list.v.list[2].type = TYPE_NUM;
    list.v.list[2].v.num = -1;		/* conventional value */
    list.v.list[3].type = TYPE_NUM;
    list.v.list[3].v.num = 0;		/* OBSOLETE: was clock ID */
    list.v.list[4].type = TYPE_NUM;
    list.v.list[4].v.num = DEFAULT_BG_TICKS;	/* OBSOLETE: was clock ticks */

    list = list_for_vm(list, the_vm);

    list.v.list[5].v.obj = player;

    return list;
}

enum error
kill_task(int id, Objid owner)
{ 	/* works for FORKED, READING and SUSPENDED tasks */
    task      **tt;
    tqueue     *tq;

    if (id == current_task_id) {
	abort_running_task();
	return E_NONE;
    }

    for (tt = &waiting_tasks; *tt; tt = &((*tt)->next)) {
	task   *t = *tt;
	Objid	progr;
	
	if (t->kind == TASK_FORKED  &&  t->t.forked.id == id)
	    progr = t->t.forked.a.progr;
	else if (t->kind == TASK_SUSPENDED  &&  t->t.suspended.id == id)
	    progr = progr_of_cur_verb(t->t.suspended.the_vm);
	else
	    continue;

	if (!is_wizard(owner)  &&  owner != progr)
	    return E_PERM;
	tq = find_tqueue(progr, 0);
	if (tq) {
	    tq->num_bg_tasks--;
#if 0
	    logf("KILL_TASK: decr #%d => %d\n", tq->player, tq->num_bg_tasks);
#endif
	}
	*tt = t->next;
	free_task(t, 1);
	return E_NONE;
    }

    for (tq = idle_tqueues; tq; tq = tq->next) {
	if (tq->reading && tq->reading_task_id == id) {
	    if (!is_wizard(owner) && owner != tq->player)
		return E_PERM;
	    free_vm(tq->reading_vm, 1);
	    tq->reading = 0;
	    return E_NONE;
	}	
    }
	
    for (tq = active_tqueues; tq; tq = tq->next) {

	if (tq->reading && tq->reading_task_id == id) {
	    if (!is_wizard(owner) && owner != tq->player)
		return E_PERM;
	    free_vm(tq->reading_vm, 1);
	    tq->reading = 0;
	    return E_NONE;
	}	
	
	for (tt = &(tq->first_bg); *tt; tt = &((*tt)->next)) {
	    task       *t = *tt;
	    
	    if ((t->kind == TASK_FORKED  &&  t->t.forked.id == id)
		|| (t->kind == TASK_SUSPENDED  &&  t->t.suspended.id == id)) {
		if (!is_wizard(owner)  &&  owner != tq->player)
		    return E_PERM;
		*tt = t->next;
		if (t->next == 0)
		    tq->last_bg = tt;
		tq->num_bg_tasks--;
#if 0
		logf("KILL_TASK: decr #%d => %d\n",
		     tq->player, tq->num_bg_tasks);
#endif
		free_task(t, 1);
		return E_NONE;
	    }
	}
    }

    return E_INVARG;
}

static void
write_forked_task(forked_task ft)
{
    /* losing some info because not all of the program written */
    int lineno = get_lineno(ft.program, 0, ft.f_index);

    db_printf("0 %d %d %d\n", lineno, ft.start_time, ft.id);
    write_activ_as_pi(ft.a);
    write_rt_env(ft.program->var_names, ft.rt_env, ft.program->num_var_names);
    write_forked_program(ft.program, ft.f_index);
}

static void
write_suspended_task(suspended_task st)
{
    db_printf("%d %d\n", st.start_time, st.id);
    write_vm(st.the_vm);
}

void
write_task_queue(FILE *f)
{
    int		forked_count = 0;
    int		suspended_count = 0;
    task       *t;
    tqueue     *tq;

    db_printf("0 clocks\n");	/* for compatibility's sake */

    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_FORKED)
	    forked_count++;
	else /* t->kind == TASK_SUSPENDED */
	    suspended_count++;
    
    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first_bg; t; t = t->next)
	    if (t->kind == TASK_FORKED)
		forked_count++;
	    else /* t->kind == TASK_SUSPENDED */
		suspended_count++;
    

    db_printf("%d queued tasks\n", forked_count); /***** has to change name
						     to forked tasks someday */

    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_FORKED)
	    write_forked_task(t->t.forked);

    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first_bg; t; t = t->next)
	    if (t->kind == TASK_FORKED)
		write_forked_task(t->t.forked);

    db_printf("%d suspended tasks\n", suspended_count);
    
    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_SUSPENDED)
	    write_suspended_task(t->t.suspended);

    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first_bg; t; t = t->next)
	    if (t->kind == TASK_SUSPENDED)
		write_suspended_task(t->t.suspended);

}

int
read_task_queue(FILE *f)
{
    int		count, suspended_count, suspended_task_header;

    /* Skip obsolete clock stuff */
    if (fscanf(f, "%d clocks\n", &count) != 1) {
        errlogf("READ_TASK_QUEUE: Bad clock count.\n");
        return -1;
    }
    for (; count > 0; count--) {
	int	dummy;

	if (fscanf(f, "%d %d %d\n", &dummy, &dummy, &dummy) != 3) {
            errlogf("READ_TASK_QUEUE: Bad clock; count = %d\n", count);
            return -1;
        }
    }

    if (fscanf(f, "%d queued tasks\n", &count) != 1) {
        errlogf("READ_TASK_QUEUE: Bad task count.\n");
        return -1;
    }
    for (; count > 0; count--) {
	int		i, first_lineno, id, old_size, st;
	time_t		start_time;
	Program	       *program;
	Var	       *rt_env, *old_rt_env;
	const char    **old_names;
	activation      a;

	if (fscanf(f, "%d %d %d %d", &i, &first_lineno, &st, &id) != 4
	    || fgetc(f) != '\n') {
	    errlogf("READ_TASK_QUEUE: Bad numbers, count = %d.\n", count);
            return -1;
        }
	start_time = st;
	if (read_activ_as_pi(f, &a) < 0) {
            errlogf("READ_TASK_QUEUE: Bad activation, count = %d.\n", count);
            return -1;
        }
        if (read_rt_env(f, &old_names, &old_rt_env, &old_size) < 0) {
            errlogf("READ_TASK_QUEUE: Bad env, count = %d.\n", count);
            return -1;
        }
        if (read_program(f, &program) < 0) {
            errlogf("READ_TASK_QUEUE: Bad program, count = %d.\n", count);
            return -1;
        }
	rt_env = reorder_rt_env(old_rt_env, old_names, old_size, program);
	program->first_lineno = first_lineno;
	
	enqueue_ft(program, a, rt_env, MAIN_VECTOR, start_time, id);
	/***** wierd! the forked statement is parsed as the main vector
	  hence this is a forked task but running on the MAIN vector */
    }	

    suspended_task_header = fscanf(f, "%d suspended tasks\n", &suspended_count);
    if (suspended_task_header == EOF)
	return 0; /* old version */

    if (suspended_task_header != 1) {
	errlogf("READ_TASK_QUEUE: Bad suspended task count.\n");
	return -1;
    }

    for (; suspended_count > 0; suspended_count--) {
	task   *t = (task *) mymalloc(sizeof(task), M_TASK);
	int	st;

	t->kind = TASK_SUSPENDED;
	if (fscanf(f, "%d %d\n", &st, &(t->t.suspended.id)) != 2) {
	    errlogf("READ_TASK_QUEUE: Bad suspended task header. count = %d\n",
		   suspended_count);
	    return -1;
	}
	t->t.suspended.start_time = st;

	if (read_vm(f, &(t->t.suspended.the_vm)) < 0) {
	    errlogf("READ_TASK_QUEUE: Bad suspended task vm. count = %d\n",
		   suspended_count);
	    return -1;
	}

	enqueue_waiting(t);
    }
    return 0;
}

/**** built-in functions ****/

static package
bf_queue_info(Var arglist, Byte next, void *vdata, Objid progr)
{
    int		nargs = arglist.v.list[0].v.num;
    Var		res;

    if (nargs == 0) {
	int	count = 0;
	tqueue *tq;

	for (tq = active_tqueues; tq; tq = tq->next)
	    count++;
	for (tq = idle_tqueues; tq; tq = tq->next)
	    count++;

	res = new_list(count);
	for (tq = active_tqueues; tq; tq = tq->next) {
	    res.v.list[count].type = TYPE_OBJ;
	    res.v.list[count].v.obj = tq->player;
	    count--;
	}
	for (tq = idle_tqueues; tq; tq = tq->next) {
	    res.v.list[count].type = TYPE_OBJ;
	    res.v.list[count].v.obj = tq->player;
	    count--;
	}
    } else {
	Objid	who = arglist.v.list[1].v.obj;
	tqueue *tq = find_tqueue(who, 0);

	res.type = TYPE_NUM;
	res.v.num = (tq ? tq->num_bg_tasks : 0);
    }

    free_var(arglist);
    return make_var_pack(res);
}

static package
bf_task_id(Var arglist, Byte next, void *vdata, Objid progr) 
{ /* () */
    Var r;
    r.type = TYPE_NUM;
    r.v.num = current_task_id;
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_queued_tasks(Var arglist, Byte next, void *vdata, Objid progr)
{ /* lists FORKED, READING and SUSPENDED tasks */   
    Var		tasks;
    int		show_all = is_wizard(progr);
    tqueue     *tq;
    task       *t;
    int		i, count = 0;

    for (tq = idle_tqueues; tq; tq = tq->next) {
	if (tq->reading &&  (show_all  ||  tq->player == progr))
	    count++;
    }

    for (tq = active_tqueues; tq; tq = tq->next) {
	if (tq->reading &&  (show_all  ||  tq->player == progr))
	    count++;

	for (t = tq->first_bg; t; t = t->next)
	    if (show_all
		|| (t->kind == TASK_FORKED
		    ? t->t.forked.a.progr == progr
		    : progr_of_cur_verb(t->t.suspended.the_vm) == progr))
		count++;
    }		    

    for (t = waiting_tasks; t; t = t->next) 
	if (show_all
	    || (t->kind == TASK_FORKED
		? t->t.forked.a.progr == progr
		: progr_of_cur_verb(t->t.suspended.the_vm) == progr))
	    count++;
    
    tasks = new_list(count);
    i = 1;

    for (tq = idle_tqueues; tq; tq = tq->next) {
	if (tq->reading && (show_all  ||  tq->player == progr))
	    tasks.v.list[i++] = list_for_reading_task(tq->player,
						      tq->reading_task_id,
						      tq->reading_vm);
    }

    for (tq = active_tqueues; tq; tq = tq->next) {
	if (tq->reading && (show_all  ||  tq->player == progr))
	    tasks.v.list[i++] = list_for_reading_task(tq->player,
						      tq->reading_task_id,
						      tq->reading_vm);

	for (t = tq->first_bg; t; t = t->next)
	    if (t->kind == TASK_FORKED  &&  (show_all  ||
					     t->t.forked.a.progr == progr))
		tasks.v.list[i++] = list_for_forked_task(t->t.forked);
	    else if (t->kind == TASK_SUSPENDED
		     && (show_all
			 || progr_of_cur_verb(t->t.suspended.the_vm) == progr))
		tasks.v.list[i++] = list_for_suspended_task(t->t.suspended);
    }		    

    for (t = waiting_tasks; t; t = t->next) {
	if (t->kind == TASK_FORKED  &&  (show_all  ||
					 t->t.forked.a.progr == progr))
	    tasks.v.list[i++] = list_for_forked_task(t->t.forked);
	else if (t->kind == TASK_SUSPENDED
		 && (show_all
		     || progr_of_cur_verb(t->t.suspended.the_vm) == progr))
	    tasks.v.list[i++] = list_for_suspended_task(t->t.suspended);
    }

    free_var(arglist);
    return make_var_pack(tasks);
}

static package
bf_kill_task(Var arglist, Byte next, void *vdata, Objid progr)
{
    enum error	e = kill_task(arglist.v.list[1].v.num, progr);
    free_var(arglist);
    
    if (e != E_NONE)
	return make_error_pack(e);
    
    return no_var_pack();
}

static package
bf_output_delimiters(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    Objid player = arglist.v.list[1].v.obj;
    
    free_var(arglist);
    
    if (!is_wizard(progr) && progr != player) 
	return make_error_pack(E_PERM);
    else {
	const char *prefix, *suffix;
	tqueue     *tq = find_tqueue(player, 0);

	if (!tq || !tq->connected)
	    return make_error_pack(E_INVARG);

	if (tq->output_prefix)
	    prefix = tq->output_prefix;
	else
	    prefix = "";

	if (tq->output_suffix)
	    suffix = tq->output_suffix;
	else
	    suffix = "";

	r = new_list(2);
	r.v.list[1].type = r.v.list[2].type = TYPE_STR;
	r.v.list[1].v.str = str_dup(prefix);
	r.v.list[2].v.str = str_dup(suffix);
    }
    return make_var_pack(r);
}

void
register_tasks(void)
{ 
    register_function("task_id", 0, 0, bf_task_id);
    register_function("queued_tasks", 0, 0, bf_queued_tasks);
    register_function("kill_task", 1, 1, bf_kill_task, TYPE_NUM);
    register_function("output_delimiters", 1, 1, bf_output_delimiters,
		      TYPE_OBJ);
    register_function("queue_info", 0, 1, bf_queue_info, TYPE_OBJ);
}

char rcsid_tasks[] = "$Id: tasks.c,v 1.15 1992/10/23 23:03:47 pavel Exp $";

/* $Log: tasks.c,v $
 * Revision 1.15  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.14  1992/10/23  22:22:38  pavel
 * Eliminated all uses of the useless macro NULL.
 *
 * Revision 1.13  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.12  1992/10/17  20:54:19  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.11  1992/09/14  17:40:22  pjames
 * Moved db_modification code to db modules.
 *
 * Revision 1.10  1992/09/08  21:57:11  pjames
 * Inserted all procedures from bf_task.c
 *
 * Revision 1.9  1992/09/02  18:39:49  pavel
 * Fixed bug whereby read()ing tasks are never resumed if the connection being
 * read is dropped while they're waiting for new input.
 *
 * Revision 1.8  1992/08/31  22:24:03  pjames
 * Changed some `char *'s to `const char *'
 *
 * Revision 1.7  1992/08/28  16:06:01  pjames
 * Changed myfree(*, M_STRING) to free_str(*).
 * Changed some strref's to strdup.
 *
 * Revision 1.6  1992/08/21  00:39:11  pavel
 * Renamed include file "parse_command.h" to "parse_cmd.h".
 *
 * Added output to player on each use of `.program' declaring that command to
 * be obsolete.
 *
 * Added check for `out of band' commands, which bypass both normal command
 * parsing and any read()ing task to become server tasks invoking
 * #0:do_out_of_band_command(@word-list), for use by fancy clients to
 * communicate with the server despite the player's current state.
 *
 * Revision 1.5  1992/08/12  01:51:16  pjames
 * Copied verb and verbname when creating a forked_task.
 *
 * Revision 1.4  1992/08/10  16:50:03  pjames
 * Updated #includes.  Changed forked_task data structure to hold an
 * activation instead of a Parse_Info, and all manipulation of
 * forked_task.pi now uses correct fields from forked_task.a;
 *
 * Revision 1.3  1992/07/27  18:24:33  pjames
 * Change ct_env_size to num_var_names.
 *
 * Revision 1.2  1992/07/21  00:07:17  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
